
/* Smart UI v15.1.2 (2023-04-02) 
Copyright (c) 2011-2023 jQWidgets. 
License: https://htmlelements.com/license/ */ //

Smart('smart-input', class Input extends Smart.BaseElement {
    // Button's properties.
    static get properties() {
        return {
            'autoCompleteDelay': {
                value: 100,
                type: 'number'
            },
            'dataSource': {
                type: 'any',
                value: [],
                reflectToAttribute: false
            },
            'dropDownClassList': {
                value: [],
                type: 'array',
                reflectToAttribute: false
            },
            'dropDownDataSource': {
                type: 'any',
                value: []
            },
            'dropDownHeight': {
                type: 'any',
                value: 200
            },
            'dropDownOpenMode': {
                allowedValues: ['default', 'manual'],
                value: 'default',
                type: 'string'
            },
            'dropDownAppendTo': {
                value: null,
                type: 'any'
            },
            'dropDownWidth': {
                type: 'any',
                value: null
            },
            'dropDownButtonPosition': {
                allowedValues: ['none', 'left', 'right'],
                value: 'none',
                type: 'string'
            },
            'dropDownOpenPosition': {
                allowedValues: ['bottom', 'top', 'auto'],
                value: 'bottom',
                type: 'string'
            },
            'hint': {
                type: 'string',
                value: ''
            },
            'inputPurpose': {
                type: 'string',
                value: 'off'
            },
            'items': {
                type: 'number',
                value: 8
            },
            'minLength': {
                type: 'number',
                value: 1
            },
            'maxLength': {
                type: 'number',
                value: 99999999
            },
            'name': {
                value: '',
                type: 'string'
            },
            'prefix': {
                value: '',
                type: 'string',
                reflectToAttribute: false
            },
            'suffix': {
                value: '',
                type: 'string',
                reflectToAttribute: false
            },
            'onAffixClick': {
                value: null,
                type: 'any',
                reflectToAttribute: false
            },
            'opened': {
                value: false,
                type: 'boolean'
            },
            'query': {
                type: 'string',
                value: ''
            },
            'queryMode': {
                allowedValues: ['contains', 'containsIgnoreCase', 'doesNotContain', 'doesNotContainIgnoreCase', 'equals', 'equalsIgnoreCase', 'startsWith', 'startsWithIgnoreCase', 'endsWith', 'endsWithIgnoreCase'],
                type: 'string',
                value: 'containsIgnoreCase'
            },
            'placeholder': {
                value: '',
                type: 'string'
            },
            'readonly': {
                type: 'boolean',
                value: false
            },
            'sorted': {
                value: false,
                type: 'boolean'
            },
            'sortDirection': {
                value: 'asc',
                type: 'string'
            },
            'selectedIndex': {
                type: 'number',
                value: -1
            },
            'selectedValue': {
                type: 'any',
                value: ''
            },
            'selectedValues': {
                value: [],
                type: 'array'
            },
            'type': {
                type: 'string'
            },
            'value': {
                type: 'string',
                value: ''
            },
            'innerHTML': {
                type: 'string',
                reflectToAttribute: false
            }
        };
    }

    /**
    * CSS files needed for the element (ShadowDOM)
    */
    static get styleUrls() {
        return [
            'smart.textbox.css'
        ]
    }

    /** Input's template. */
    template() {
        return '<div id="inputContainer" role="presentation"><div id="optionsContainer" inner-h-t-m-l=\'{{innerHTML}}\' class="smart-hidden"></div><span class="smart-input-prefix" id="prefix" inner-h-t-m-l=\'{{prefix}}\'></span><input maxlength=\'[[maxLength]]\'  class="smart-input" id=\'input\' readonly=\'[[readonly]]\' placeholder=\'[[placeholder]]\' type=\'[[type]]\' name=\'[[name]]\' value=\'{{value::keyup}}\' disabled=\'[[disabled]]\' aria-label="[[placeholder]]" /><span class="smart-hidden smart-hint" id="span">[[hint]]</span><span id="suffix"  class="smart-input-suffix" inner-h-t-m-l=\'{{suffix}}\'></span><div id="dropDownButton" tabindex=-1 class="smart-drop-down-button" role="button" aria-label="Toggle popup"><div id="arrow" class="arrow" aria-hidden="true"></div></div></div>';
    }

    static get listeners() {
        return {
            'input.focus': '_focusHandler',
            'input.blur': '_blurHandler',
            'input.keydown': '_keyDownHandler',
            'input.keyup': '_keyUpHandler',
            'input.keypress': '_keyPressHandler',
            'input.paste': '_pasteHandler',
            'prefix.click': '_prefixClick',
            'suffix.click': '_suffixClick',
            'dropDownButton.down': '_dropDownButtonDownHandler',
            'inputContainer.down': '_downHandler',
            'document.up': '_documentUpHandler'
        };
    }

    _prefixClick(event) {
        const that = this;


        if (that.onAffixClick) {
            if (typeof that.onAffixClick === 'string') {
                const fnName = that.onAffixClick.replace('()', '');
                if (window[fnName]) {
                    window[fnName].apply(that, event);
                }
            }
            else {
                that.onAffixClick(event);
            }
        }
    }

    _suffixClick(event) {
        const that = this;
        if (typeof that.onAffixClick === 'string') {
            const fnName = that.onAffixClick.replace('()', '');
            if (window[fnName]) {
                window[fnName].apply(that, event);
            }
        }
        else {
            that.onAffixClick(event);
        }
    }

    get selectedIndexes() {
        if (!this.isCompleted) {
            return [];
        }

        return [this.selectedIndex];
    }

    set selectedIndexes(value) {
        if (!this.isCompleted) {
            return;
        }

        const that = this;

        if (value && value[0] !== undefined) {
            this.selectedIndex = value[0];

            if (this.dataSource === null) {
                return;
            }
            const item = this.dataSource[this.selectedIndex];

            if (!item) {
                return;
            }

            const context = this.context;
            this.context = this;
            this.$.input.dataValue = this.selectedValue = typeof item === 'string' ? item : item.value;
            this.setAttribute('data-value', this.$.input.dataValue);

            let items = that.dataSource;
            if (typeof items === 'string') {
                items = that.$.deserialize(items, 'array');
            }

            for (let i = 0; i < items.length; i++) {
                let item = items[i];
                if (typeof item === 'string' || typeof item === 'number') {
                    continue;
                }

                if (item.value === that.$.input.dataValue) {
                    item.selected = true;
                }
                else {
                    item.selected = false;
                }
            }
            this.context = context;
        }
        else {
            const context = this.context;
            this.context = this;
            this.$.input.dataValue = this.selectedValue = '';
            this.setAttribute('data-value', this.$.input.dataValue);
            this.selectedIndex = -1;
            let items = that.dataSource;
            if (typeof items === 'string') {
                items = that.$.deserialize(items, 'array');
            }

            if (items && Array.isArray(items)) {
                for (let i = 0; i < items.length; i++) {
                    let item = items[i];
                    if (typeof item === 'string') {
                        continue;
                    }

                    if (item.value === that.$.input.dataValue) {
                        item.selected = true;
                    }
                    else {
                        item.selected = false;
                    }
                }
            }
            this.context = context;
        }
    }

    focus() {
        const that = this;

        that.$.input.focus();
    }

    select() {
        const that = this;

        if (!that.readonly) {
            that.$.input.select();
        }
        else {
            that.$.input.focus();
        }
    }

    _documentUpHandler(event) {
        const that = this;

        const target = that.shadowRoot || that.isInShadowDOM ? event.originalEvent.composedPath()[0] : event.originalEvent.target;

        if (target === that ||
            target === that.$.dropDownButton ||
            target === that.$.input ||
            target === that.$.arrow) {
            that._isPointerDown = false;
            return;
        }

        if (that.$.scrollView.contains(target.shadowParent || target)) {
            if (that._isPointerDown) {
                that._isPointerDown = false;

                if (that.opened) {
                    that._preventLookup = true;
                }

                that.$.input.focus();
            }

            return;
        }

        if (that.opened) {
            that._preventLookup = true;
        }

        if (that.dropDownOpenMode !== 'manual' && !that._isPointerDown) {
            that.close();
        }

        that._isPointerDown = false;
    }

    _focusHandler() {
        const that = this;

        that.setAttribute('focus', '');

        if (!that.readonly && that.minLength === 0 && that.$.input.value.length === 0 && !that._preventLookup) {
            that._lookup();
            return;
        }

        that.$.fireEvent('focus');

        delete that._preventLookup;
    }

    _blurHandler() {
        const that = this;

        if (!that.opened) {
            that.removeAttribute('focus');
        }


        if (!that.opened) {
            that.$.fireEvent('blur');
        }

        that._refreshSelectedIndex();

        delete that._preventLookup;
    }

    _pasteHandler(event) {
        const that = this,
            input = that.$.input,
            inputValue = input.value,
            selectionStart = input.selectionStart,
            selectionEnd = input.selectionEnd;

        const hasClipboardData = (event.clipboardData || window.clipboardData);

        if (!hasClipboardData) {
            return;
        }

        const clipboardData = (event.clipboardData || window.clipboardData).getData('text');
        let newValue = inputValue.slice(0, selectionStart) + clipboardData,
            carretPosition = newValue.length;

        newValue += inputValue.slice(selectionEnd);

        event.preventDefault();

        that.set('value', newValue);
        that.setAttribute('data-value', newValue);

        input.setSelectionRange(carretPosition, carretPosition);

        that._refreshSelectedIndex();
    }

    _refreshSelectedIndex() {
        const that = this;
        let items = that.dataSource;

        if (items) {
            let selectedIndex = -1;
            let selectedValue = null;

            if (that.dataSource instanceof Promise) {
                return;
            }

            if (typeof that.dataSource === 'function') {
                return;
            }

            if (typeof items === 'string') {
                items = that.$.deserialize(items, 'array');
            }

            for (let i = 0; i < items.length; i++) {
                let item = items[i];

                if (typeof item === 'number' || typeof item === 'boolean') {
                    item = '' + item;
                }

                if (typeof item === 'string') {
                    if (item === that.value) {
                        selectedIndex = i;
                        selectedValue = that.value;
                        break;
                    }
                }
                else if (item) {
                    delete item.selected;

                    if (item.label === that.value) {
                        selectedIndex = i;
                        selectedValue = item.value;
                        item.selected = true;
                    }
                }
            }

            that.set('selectedIndex', selectedIndex, false);
            that.set('selectedValue', selectedValue, false);

            if (that.selectedIndex === -1) {
                that.$.input.dataValue = null;
            }
        }
    }

    selectItem(value) {
        const that = this;

        let items = that.dataSource;

        if (typeof items === 'string') {
            items = that.$.deserialize(items, 'array');
        }

        if (items) {
            for (let i = 0; i < items.length; i++) {
                const item = items[i];

                that.set('selectedIndex', -1);
                that.set('value', '');
                that.$.input.dataValue = null;

                if (typeof item === 'string') {
                    if (item === value) {
                        that.set('selectedIndex', i, false);
                        that.set('value', that.$.input.dataValue = value);
                        that.setAttribute('data-value', that.$.input.dataValue);
                        return;
                    }
                }
                else {
                    if (item.value === value) {
                        that.set('selectedIndex', i, false);
                        that.set('value', item.label);
                        that.$.input.dataValue = value;
                        that.setAttribute('data-value', that.$.input.dataValue);
                        return;
                    }
                }
            }
        }
        else {
            that.set('value', that.$.input.dataValue = value);
            that.setAttribute('data-value', that.$.input.dataValue);
            that.set('selectedIndex', -1, false);
        }
    }

    getItem(value) {
        const that = this;

        if (that.dataSource) {
            for (let i = 0; i < that.dataSource.length; i++) {
                const item = that.dataSource[i];

                if (item.value === value) {
                    return item;
                }
            }
        }

        return null;
    }

    getPrevItem() {
        const that = this;

        const selectedItem = that.getSelectedItem();

        if (that.dataSource && selectedItem) {
            for (let i = 0; i < that.dataSource.length; i++) {
                const item = that.dataSource[i];

                if (item.value === selectedItem.value) {
                    const prevItem = that.dataSource[i - 1];

                    return prevItem;
                }
            }
        }

        return null;
    }

    getNextItem() {
        const that = this;

        const selectedItem = that.getSelectedItem();

        if (that.dataSource && selectedItem) {
            for (let i = 0; i < that.dataSource.length; i++) {
                const item = that.dataSource[i];

                if (item.value === selectedItem.value) {
                    const nextItem = that.dataSource[i + 1];

                    return nextItem;
                }
            }
        }

        return null;
    }

    getSelectedItem() {
        const that = this;

        if (that.dataSource && that.dataSource.length === 0) {
            return { label: that.value, value: that.$.input.dataValue ? that.$.input.dataValue : that.value }
        }

        return { label: that.value, value: that.$.input.dataValue !== undefined ? that.$.input.dataValue : that.value };
    }

    _refreshMenu() {
        const that = this;

        that.$.menu.classList.remove('scroll');

        if (that.$.scrollView.computedVerticalScrollBarVisibility) {
            that.$.menu.classList.add('scroll');
        }
    }

    _performSelect() {
        const that = this;
        const targetItem = that.$.menu.querySelector('.active');
        let label = '';
        let value = '';

        if (targetItem) {
            label = targetItem.getAttribute('data-label');
            value = targetItem.dataValue;
        }

        let oldLabel = that.value,
            oldValue = that.$.input.dataValue;

        that.set('value', label);
        that.$.input.dataValue = value;
        that.setAttribute('data-value', value);
        if (targetItem) {
            that.set('selectedIndex', targetItem.index);
        }
        that.set('selectedValues', [that.$.input.dataValue]);
        that.set('selectedValue', that.$.input.dataValue);

        let items = that.dataSource;
        if (typeof items === 'string') {
            items = that.$.deserialize(items, 'array');
        }

        if (items && Array.isArray(items)) {
            for (let i = 0; i < items.length; i++) {
                let item = items[i];
                if (typeof item === 'string' || typeof item === 'number') {
                    continue;
                }

                if (item.value === that.$.input.dataValue) {
                    item.selected = true;
                }
                else {
                    item.selected = false;
                }
            }
        }
        that.close();

        if (label !== oldLabel || value !== oldValue) {
            that.$.fireEvent('change', { value: value, label: label, oldValue: oldValue, oldLabel: oldLabel });
            that._updateTargetValue();
        }
    }

    _open() {
        const that = this;

        if (that.opened) {
            return;
        }

        const scrollX = window.scrollX,
            scrollY = window.scrollY;
        let rect = that.getBoundingClientRect(),
            xCorrection = 0,
            yCorrection = 0;

        if (that.timer) {
            clearTimeout(that.timer);
        }

        if (that.getRootNode() && that.getRootNode().host) {
            that.getRootNode().host.shadowRoot.appendChild(that.$.scrollView);
        }
        else {
            if (!that.dropDownAppendTo) {
                document.body.appendChild(that.$.scrollView);
            }
            else {
                if (that.dropDownAppendTo === 'body' || that.dropDownAppendTo === document.body) {
                    document.body.appendChild(that.$.scrollView);
                }
                else if (that.dropDownAppendTo instanceof HTMLElement) {
                    that.dropDownAppendTo.appendChild(that.$.scrollView);
                }
                else if (typeof dropDownAppendTo === 'string') {
                    let dropDownAppendTo = document.getElementById(that.dropDownAppendTo);

                    if (dropDownAppendTo instanceof HTMLElement) {
                        dropDownAppendTo.appendChild(that.$.scrollView);
                    }
                    else {
                        document.body.appendChild(that.$.scrollView);
                    }
                }
            }
        }

        that.setAttribute('aria-owns', that.$.scrollView.id);

        if (!that.readonly) {
            that.$.input.setAttribute('aria-controls', that.$.scrollView.id);
        }

        if (that.$.scrollView.enableShadowDOM && !that._importedStyle) {
            that.$.scrollView.importStyle(that._getStyleUrl('smart.textbox.css'));
            that._importedStyle = true;
        }

        if (getComputedStyle(document.body).position !== 'static') {
            const bodyRect = document.body.getBoundingClientRect();

            xCorrection = bodyRect.left;
            yCorrection = bodyRect.top;
        }

        that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', '');
        that.$.scrollView.style.left = -3 + rect.left + scrollX - xCorrection + 'px';
        that.$.scrollView.classList.remove('open');
        that.$.scrollView.removeAttribute('top');

        if (that.dropDownOpenPosition === 'bottom') {
            that.$.scrollView.style.top = rect.bottom + scrollY - yCorrection + 1 + 'px';
        }
        else if (that.dropDownOpenPosition === 'top') {
            that.$.scrollView.style.top = rect.top - that.$.scrollView.offsetHeight + scrollY - yCorrection - 2 + 'px';
            that.$.scrollView.setAttribute('top', '');
        }
        else if (that.dropDownOpenPosition === 'auto') {
            that.$.scrollView.style.top = rect.bottom + scrollY - yCorrection + 1 + 'px';
            const topOpenPosition = rect.top - that.$.scrollView.offsetHeight + scrollY - yCorrection - 2;
            if (window.innerHeight < parseInt(that.$.scrollView.style.top) + that.$.scrollView.offsetHeight - scrollY) {
                if (topOpenPosition >= 0) {
                    that.$.scrollView.style.top = topOpenPosition + 'px';
                    that.$.scrollView.setAttribute('top', '');
                }
            }
            else {
                that.$.scrollView.style.top = rect.bottom + scrollY - yCorrection + 1 + 'px';
            }
        }

        if (that.dropDownAppendTo && that.dropDownAppendTo !== 'body') {
            that.$.scrollView.style.position = 'absolute';
            that.$.scrollView.style.width = 'auto';
            that.$.scrollView.style.height = 'auto';
            const parent = that.$.scrollView.parentNode;
            const parentRect = parent.getBoundingClientRect();

            that.$.scrollView.style.top = rect.bottom - parentRect.top + 'px';
            if (that.dropDownOpenPosition === 'top') {
                that.$.scrollView.style.top = rect.top - parentRect.top - that.$.scrollView.offsetHeight + 'px';
            }

            that.$.scrollView.style.left = -3 + rect.left - parentRect.left + 'px';

            if (that.dropDownHeight) {
                that.$.scrollView.style.height = that.dropDownHeight;
            }
            if (that.dropDownWidth) {
                that.$.scrollView.style.width = that.dropDownWidth;
            }
        }

        that.opening = true;

        if (that._positionTimer) {
            clearInterval(that._positionTimer);
        }

        if (!that.dropDownAppendTo) {
            that._positionTimer = setInterval(() => {
                const newRect = that.getBoundingClientRect();

                if (!that.opening) {
                    if (Math.abs(parseInt(newRect.top) - parseInt(rect.top)) >= 10 || Math.abs(parseInt(newRect.left) - parseInt(rect.left)) >= 10) {
                        const scrollX = window.scrollX,
                            scrollY = window.scrollY;
                        that.$.scrollView.style.left = -3 + newRect.left + scrollX - xCorrection + 'px';

                        if (that.dropDownOpenPosition === 'bottom') {
                            that.$.scrollView.style.top = newRect.bottom + scrollY - yCorrection + 1 + 'px';
                        }
                        else if (that.dropDownOpenPosition === 'top') {
                            that.$.scrollView.style.top = newRect.top - that.$.scrollView.offsetHeight + scrollY - yCorrection + 1 + 'px';
                        }
                        else if (that.dropDownOpenPosition === 'auto') {
                            that.$.scrollView.removeAttribute('top');
                            that.$.scrollView.style.top = newRect.bottom + scrollY - yCorrection + 1 + 'px';

                            if (window.innerHeight < parseInt(that.$.scrollView.style.top) + that.$.scrollView.offsetHeight - scrollY) {
                                that.$.scrollView.style.top = newRect.top - that.$.scrollView.offsetHeight + scrollY - yCorrection + 1 + 'px';
                                that.$.scrollView.setAttribute('top', '');
                            }
                            else {
                                that.$.scrollView.style.top = newRect.bottom + scrollY - yCorrection + 1 + 'px';
                            }
                        }
                        return;
                    }
                }
                else {
                    that.opening = false;
                    rect = newRect;
                }
            }, 50);
        }

        that.$.scrollView.onpointerdown = function () {
            that._isPointerDown = true;
        }

        requestAnimationFrame(function () {
            const dropDownWidth = that.dropDownWidth;

            if (that.$.scrollView.refresh) {
                that.$.scrollView.refresh();
            }

            that._refreshMenu();
            that.$.scrollView.setAttribute('open', '');
            that.setAttribute('open', '');
            that.$.dropDownButton.setAttribute('open', '');
            that.$.input.setAttribute('open', '');

            if (dropDownWidth && typeof dropDownWidth === 'string' && dropDownWidth.indexOf('%') !== -1) {
                const fraction = parseFloat(dropDownWidth) / 100;

                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', that.offsetWidth * fraction + 'px');
            }
            else if (dropDownWidth !== 'auto' && dropDownWidth) {
                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', parseFloat(dropDownWidth) + 'px');
            }
            else if (dropDownWidth === 'auto') {
                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', '');

                if (that.$.scrollView.$.scrollViewerContainer.classList.contains('vscroll')) {
                    // compensates for scrollbar's width
                    const autoWidth = that.$.scrollView.offsetWidth + parseFloat(getComputedStyle(that.$.scrollView).getPropertyValue('--smart-scroll-bar-size'));

                    that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', autoWidth + 'px');
                }
            }
            else {
                that.$.menu.style.width = 'auto';

                let menuWidth = that.$.menu.offsetWidth;

                const listItems = that.$.menu.querySelectorAll('li');

                for (let i = 0; i < listItems.length; i++) {
                    that._maxDropDownWidth = Math.max((listItems[i].firstElementChild || listItems[i]).offsetWidth, that._maxDropDownWidth);
                }

                if (that.$.scrollView.computedVerticalScrollBarVisibility) {
                    menuWidth += that.$.scrollView.$ && that.$.scrollView.$.verticalScrollBar ? that.$.scrollView.$.verticalScrollBar.offsetWidth : 30;
                }

                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-width', Math.max(menuWidth, that.offsetWidth - 8) + 'px');

                that.$.menu.style.width = '';
            }

            //ShadowDOM ScrollBar styles are not loaded when the method is invoked
            if ((that.shadowRoot || that.isInShadowDOM) && that.$.scrollView._scrollView && Smart.ScrollBar) {
                requestAnimationFrame(() => {
                    if (that.$.scrollView._scrollView) {
                        that.$.scrollView._scrollView.vScrollBar.refresh();
                    }
                });
            }
        });


        if (that.$.scrollView.refresh) {
            that.$.scrollView.refresh();
        }

        that._refreshMenu();

        that.set('opened', true);
        that.setAttribute('aria-expanded', true);
        that.$.fireEvent('open');
    }

    close() {
        const that = this;

        if (!that.opened) {
            return false;
        }

        if (that.timer) {
            clearTimeout(that.timer);
        }

        that.timer = setTimeout(function () {
            if (that.$.scrollView.parentNode && !that.opened) {
                that.$.scrollView.remove();

                that.removeAttribute('aria-owns');

                if (!that.readonly) {
                    that.$.input.removeAttribute('aria-controls');
                }
            }
        }, 1000);

        if (that._positionTimer) {
            clearInterval(that._positionTimer);
        }

        that.$.scrollView.removeAttribute('open');
        that.$.dropDownButton.removeAttribute('open');
        that.$.input.removeAttribute('open');
        that.removeAttribute('open');

        that.set('opened', false);
        that.setAttribute('aria-expanded', false);
        that.$.input.focus();
        that.$.fireEvent('close');

        return true;
    }

    /**
    * Refreshes the UI Component.
    */
    refresh() {
        const that = this;

        if (that.$.scrollView.refresh) {
            that.$.scrollView.refresh();
        }

        that._getDataSourceFromOption();

        that.close();
    }

    _lookup(event) {
        const that = this;

        that.query = that.$.input.value;

        if (that.$.input.readonly) {
            if (!that._incrementalSearchQuery) {
                that._incrementalSearchQuery = '';
            }

            that._incrementalSearchQuery += event.key;

            if (that._incrementalSearchTimer) {
                clearTimeout(that._incrementalSearchTimer);
            }

            that.query = that._incrementalSearchQuery;
            that._incrementalSearchTimer = setTimeout(function () {
                that._incrementalSearchQuery = '';
            }, 700);
        }

        if (that.query.length < that.minLength) {
            that.close();
            return;
        }

        const autoComplete = function (items) {
            clearTimeout(that._autoCompleteTimeout);

            that._autoCompleteTimeout = setTimeout(() => {
                const oldContext = that.context;

                that.context = that;
                that._process(items);
                that.context = oldContext;
            }, that.autoCompleteDelay);
        }

        typeof that.dataSource === 'function' ? that.dataSource(that.query, autoComplete) : autoComplete(that.dataSource);
    }

    _downHandler(event) {
        const that = this;

        if (that.readonly) {
            that._dropDownButtonDownHandler(event);
        }
    }

    _dropDownButtonDownHandler(event) {
        const that = this;

        if (that.dropDownOpenMode !== 'manual') {
            that._toggle();
        }

        event.preventDefault();
        event.stopPropagation();

        event.originalEvent.preventDefault();
        event.originalEvent.stopPropagation();

        that._isPointerDown = true;

        return false;
    }

    _toggle() {
        const that = this;

        if (that.opened) {
            that.close();
        }
        else {
            that.open();
        }
    }

    open() {
        const that = this;

        function processItems(items) {
            that._process(items);

            const active = that.$.menu.querySelector('.active');

            if (active) {
                const dataValue = that.$.input.dataValue;

                for (let i = 0; i < that.$.menu.children.length; i++) {
                    const item = that.$.menu.children[i];
                    const label = item.getAttribute('data-label'),
                        value = item.dataValue;

                    if (dataValue !== undefined && value === dataValue ||
                        dataValue === undefined && label === that.$.input.value) {
                        active.classList.remove('active');
                        item.classList.add('active');
                        that._setActiveDescendant(item);
                        that.$.input.dataValue = value;
                        that.setAttribute('data-value', value);
                        break;
                    }
                }
            }

            that.ensureVisible();

            that.$.input.focus();
            setTimeout(function () {
                that.$.input.focus();
            }, 25);
        }

        if (that.dropDownDataSource && ((Array.isArray(that.dropDownDataSource) && that.dropDownDataSource.length > 0) || typeof that.dropDownDataSource === 'function')) {
            that.query = '';
            typeof that.dropDownDataSource === 'function' ? that.dropDownDataSource(that.query, processItems) : processItems(that.dropDownDataSource);
        }
        else {
            that.query = '';
            typeof that.dataSource === 'function' ? that.dataSource(that.query, processItems) : processItems(that.dataSource);
        }
    }

    _process(items) {
        const that = this;

        if (typeof items === 'string') {
            items = that.$.deserialize(items, 'array');
        }

        if (that.readonly) {
            items = items.filter(function () {
                return true;
            });
        }
        else {
            items = items.filter(function (item) {
                if (item === '') {
                    return true;
                }

                return that.matcher ? that.matcher(item) : that._matcher(item);
            })
        }

        items = that._sorter(items);

        if (that.sorted) {
            items.sort((a, b) => {
                if (a.label !== undefined) {
                    return a.label.localeCompare(b.label);
                }

                return a.localeCompare(b);
            });

            if (that.sortDirection === 'desc') {
                items.reverse();
            }
        }

        if (!items.length && that.opened) {
            that.close();
        }

        if (items.length === 0) {
            if (that.allowItemsAdd !== true) {
                return;
            }
        }

        if (that.query.length > 0) {
            that._render(items.slice(0, that.items));
        }
        else {
            that._render(items);
        }

        that._open();

        if (that.$.scrollView.refresh) {
            that.$.scrollView.refresh();
        }

        that._refreshMenu();
        that.ensureVisible();
    }

    _matcher(item) {
        const that = this;
        const text = that.query;

        if (item === null) {
            item = 'null';
        }

        item = item.label || item;
        item = '' + item;

        switch (that.queryMode) {
            case 'startsWith':
                return item.startsWith(text);

            case 'startsWithIgnoreCase':
                return item.toLowerCase().startsWith(text.toLowerCase());

            case 'doesNotContain':
                return item.indexOf(text) < 0;

            case 'doesNotContainIgnoreCase':
                return item.toLowerCase().indexOf(text.toLowerCase()) < 0;

            case 'contains':
                return item.indexOf(text) > -1;

            default:
            case 'containsIgnoreCase':
                return ~item.toLowerCase().indexOf(text.toLowerCase());

            case 'equals':
                return item.localeCompare(text) === 0;

            case 'equalsIgnoreCase':
                return (item.toLowerCase().localeCompare(text.toLowerCase()) === 0);

            case 'endsWith':
                return item.endsWith(text);

            case 'endsWithIgnoreCase':
                return item.toLowerCase().endsWith(text.toLowerCase());
        }
    }

    _sorter(items) {
        const that = this;

        let beginswith = []
            , caseSensitive = []
            , caseInsensitive = []
            , item;

        if (items && items[0] === '') {
            return items;
        }

        // eslint-disable-next-line
        while (item = items.shift()) {
            let label = item.label || item;
            label = '' + label;

            if (!label.toLowerCase().indexOf(that.query.toLowerCase())) {
                beginswith.push(item)
            }
            else if (~label.indexOf(that.query)) {
                caseSensitive.push(item)
            }
            else {
                caseInsensitive.push(item)
            }
        }

        return beginswith.concat(caseSensitive, caseInsensitive)
    }

    _highlighter(item) {
        const that = this;
        const query = that.query.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g, '\\$&');

        if (!that.query) {
            return item;
        }

        return item.replace(new RegExp('(' + query + ')', 'ig'), function ($1, match) {
            return '<strong>' + match + '</strong>'
        })
    }

    _render(items) {
        const that = this;
        let index = 0;
        let selectedIndex = -1;
        let selectedItem = null;

        const listItems = items.map(function (item) {
            let label = item,
                value = item;

            if (typeof item === 'object') {
                label = item.label;
                value = item.value !== undefined ? item.value : label;
            }

            label = '' + label;

            // sanitize labels and remove scripts.
            label = label.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/'/g, '&#39;').replace(/"/g, '&quot;');

            const listItem = document.createElement('li');
            const anchor = document.createElement('a');

            anchor.href = '#';
            listItem.id = that.id + 'Item' + Math.floor((1 + Math.random()) * 0x10000).toString(16).substring(1);
            listItem.setAttribute('data-label', label);
            listItem.index = index++;
            listItem.dataValue = value;

            if (typeof value === 'string') {
                listItem.setAttribute('value', value);
            }
            else {
                listItem.setAttribute('value', JSON.stringify(value));
            }

            listItem.setAttribute('role', 'option');
            listItem.setAttribute('aria-label', label);
            if (item.menuLabel !== undefined) {
                label = item.menuLabel;
                anchor.innerHTML = label;
            }
            else {
                anchor.innerHTML = that.highlighter ? that.highlighter(label) : that._highlighter(label);
            }

            anchor.setAttribute('aria-hidden', true);

            if (item.disabled) {
                listItem.classList.add('smart-disabled');
            }

            if (item.icon) {
                anchor.classList.add('icon');
                anchor.classList.add(item.icon);
            }

            if (item.prefix) {
                const prefix = document.createElement('span');
                prefix.innerHTML = item.prefix;

                anchor.classList.add('has-prefix');
                anchor.innerHTML = prefix.innerHTML + anchor.innerHTML;
            }

            if (item.suffix) {
                const suffix = document.createElement('span');
                suffix.innerHTML = item.suffix;

                anchor.classList.add('has-suffix');
                anchor.innerHTML = anchor.innerHTML + suffix.innerHTML;
            }

            if (item.dataset) {
                for (let dataItem in item.dataset) {
                    listItem.setAttribute('data-' + dataItem, item.dataset[dataItem]);
                }
            }

            if (item.className) {
                const classNames = item.className.split(' ');
                for (let i = 0; i < classNames.length; i++) {
                    listItem.classList.add(classNames[i].trim());
                }
            }
            if (that.selectedValue !== undefined && item.value !== undefined && that.selectedValue !== '' && that.selectedValue === item.value) {
                item.selected = true;
                selectedItem = item;
                selectedIndex = listItem.index;
                listItem.classList.add('active');
            }

            listItem.appendChild(anchor);

            return listItem;
        })

        if (selectedIndex === -1 && listItems.length > 0) {
            listItems[0].classList.add('active');
            that._setActiveDescendant(listItems[0]);
        }

        if (selectedItem) {
            that.$.input.value = selectedItem.label;
            that.$.input.dataValue = selectedItem.value;
            that.setAttribute('data-value', that.$.input.dataValue);
        }

        that.$.menu.innerHTML = '';

        function setActiveState() {
            const previouslyActive = that.$.menu.getElementsByClassName('active');

            if (previouslyActive[0]) {
                previouslyActive[0].classList.remove('active');
            }

            this.classList.add('active');
            that._setActiveDescendant(this);
        }

        for (let i = 0; i < listItems.length; i++) {
            const listItem = listItems[i];

            that.$.menu.appendChild(listItem);

            listItem.onmouseenter = setActiveState;

            listItem.onmouseleave = function () {
                this.classList.remove('active');
                that._setActiveDescendant(null);
            }

            listItem.onclick = setActiveState;
        }
    }

    ensureVisible() {
        const that = this;
        const item = that.$.menu.querySelector('.active');

        if (!item) {
            return;
        }

        if (item.offsetTop + item.offsetHeight >= that.$.scrollView.scrollTop + that.$.scrollView.offsetHeight) {
            that.$.scrollView.scrollTop = item.offsetTop + item.offsetHeight;
        }

        if (item.offsetTop <= that.$.scrollView.scrollTop || item.offsetTop >= that.$.scrollView.scrollTop + that.$.scrollView.offsetHeight) {
            that.$.scrollView.scrollTop = item.offsetTop;
        }

        if ([...that.$.menu.children].indexOf(item) === 0) {
            that.$.scrollView.scrollTop = 0;
        }
        else if ([...that.$.menu.children].indexOf(item) === that.$.menu.children.length - 1) {
            that.$.scrollView.scrollTop = that.$.scrollView.scrollHeight;
        }
    }

    _next() {
        const that = this;
        const active = that.$.menu.querySelector('.active');

        if (!active) {
            const first = that.$.menu.firstElementChild;

            first.classList.add('active');
            that._setActiveDescendant(first);
            return;
        }

        active.classList.remove('active');

        let next = active.nextElementSibling;

        if (!next) {
            next = that.$.menu.firstElementChild;
        }

        next.classList.add('active');
        that._setActiveDescendant(next);
        that.ensureVisible();
    }

    _prev() {
        const that = this;
        const active = that.$.menu.querySelector('.active');

        if (!active) {
            const first = that.$.menu.firstElementChild;

            first.classList.add('active');
            that._setActiveDescendant(first);
            return;
        }

        active.classList.remove('active');

        let prev = active.previousElementSibling;

        if (!prev) {
            prev = that.$.menu.lastElementChild;
        }

        prev.classList.add('active');
        that._setActiveDescendant(prev);
        that.ensureVisible();
    }

    _move(event) {
        const that = this;

        if (that.readonly && !that.opened) {
            if (event.key === 'ArrowUp') {
                const prevItem = that.getPrevItem();
                if (prevItem) {
                    that.selectItem(prevItem.value);
                }
                event.preventDefault()
            }
            if (event.key === 'ArrowDown') {
                const nextItem = that.getNextItem();
                if (nextItem) {
                    that.selectItem(nextItem.value);
                }
                event.preventDefault()
            }
        }

        if (!that.opened) {
            return;
        }

        switch (event.keyCode) {
            case 9: // tab
            case 13: // enter
            case 27: // escape
                event.preventDefault()
                break

            case 38: // up arrow
                event.preventDefault()
                that._prev()
                break

            case 40: // down arrow
                event.preventDefault()
                that._next()
                break
        }

        event.stopPropagation()
    }

    _keyDownHandler(event) {
        const that = this;

        that._suppressKeyPressRepeat = ![40, 38, 9, 13, 27, 16, 17, 18].includes(event.keyCode);

        that._oldValue = that.$.input.value;

        if (event.shiftKey || event.altKey || event.ctrlKey) {
            return;
        }

        that._move(event);
    }

    _keyPressHandler(event) {
        const that = this;

        if (that._suppressKeyPressRepeat) {
            return;
        }

        if (event.shiftKey || event.altKey || event.ctrlKey) {
            return;
        }

        that._move(event)
    }

    _keyUpHandler(event) {
        const that = this;

        if (event.shiftKey) {
            return;
        }

        if (event.key === 'F2') {
            return;
        }

        switch (event.keyCode) {
            case 40: // down arrow
            case 38: // up arrow
            case 16: // shift
            case 17: // ctrl
            case 18: // alt
                if (event.keyCode === 40 && event.altKey) {
                    that.open();
                }
                if (event.keyCode === 38 && event.altKey) {
                    that.close();
                }

                break;

            case 9: // tab
            case 13: // enter
                if (!that.opened) {
                    return;
                }

                that._performSelect();
                event.stopPropagation()
                event.preventDefault()
                break

            case 27: // escape
                if (!that.opened) {
                    return;
                }

                that.close();
                event.stopPropagation()
                event.preventDefault()
                break

            default:
                that._lookup(event);
                if (that.opened && !event.ctrlKey && !event.shiftKey) {
                    event.stopPropagation()
                    event.preventDefault()
                }

                that.$.fireEvent('changing', { value: that.$.input.value, oldValue: that._oldValue });

                that.$.input.dataValue = that.$.input.value;
                that.setAttribute('data-value', that.$.input.dataValue);
                that._updateTargetValue();
        }
    }

    propertyChangedHandler(propertyName, oldValue, newValue) {
        super.propertyChangedHandler(propertyName, oldValue, newValue);

        const that = this;

        if (propertyName === 'innerHTML') {
            that._getDataSourceFromOption();
            that._setSelection(false);
        }

        if (propertyName === 'selectedValue') {
            let selectedIndex = -1;

            if (that.dataSource && Array.isArray(that.dataSource)) {
                that.dataSource.forEach((item, index) => {
                    if (typeof item === 'string') {
                        if (item === newValue) {
                            selectedIndex = index;
                        }
                    }
                    else if (item && item.value === newValue) {
                        selectedIndex = index;
                    }
                });
            }

            if (selectedIndex >= 0) {
                const context = that.context;
                that.context = that;
                that.selectedIndex = selectedIndex;
                that.context = context;
            }

            that._setSelection();
        }

        if (propertyName === 'selectedIndex') {
            let items = that.dataSource;

            that.selectedValue = null;
            if (items && Array.isArray(items)) {
                for (let i = 0; i < items.length; i++) {
                    let item = items[i];
                    if (typeof item === 'string' || typeof item === 'number') {
                        continue;
                    }

                    item.selected = false;
                }
            }

            that._setSelection();
        }
        else if (propertyName === 'dataSource') {
            if (that.dataSource && Array.isArray(that.dataSource)) {
                let selectedIndex = -1;

                that.dataSource.forEach((item, index) => {
                    if (item.selected) {
                        selectedIndex = index;
                    }
                });

                if (selectedIndex >= 0) {
                    const context = that.context;
                    that.context = that;
                    that.selectedIndex = selectedIndex;
                    that.context = context;
                }
            }
            that._setSelection(false);
        }
        else if (propertyName === 'dropDownClassList') {
            const scrollView = that.$.scrollView;

            //Remove previous classes
            for (let i = 0, max = oldValue.length; i < max; i += 1) {
                scrollView.classList.remove(oldValue[i]);
            }

            //Add new classes
            for (let i = 0, max = newValue.length; i < max; i += 1) {
                scrollView.classList.add(newValue[i]);
            }
        }
        else if (propertyName === 'dropDownHeight') {
            if (that.dropDownHeight !== 'auto') {
                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-height', that.dropDownHeight + 'px');
            }
            else {
                that.$.scrollView.style.setProperty('--smart-input-drop-down-menu-height', 'auto');
            }
        }
        else if (propertyName === 'opened') {
            that.opened = oldValue;

            if (newValue) {
                that.open();
            }
            else {
                that.close();
            }
        }
        else if (propertyName === 'placeholder') {
            if (that.readonly) {
                const label = that.getAttribute('aria-label');

                if (label && label !== oldValue) {
                    return;
                }

                if (newValue) {
                    that.setAttribute('aria-label', newValue);
                }
                else {
                    that.removeAttribute('aria-label');
                }
            }
        }
        else if (propertyName === 'selectedValues') {
            if (newValue === null) {
                that.$.input.value = '';
                that.$.input.dataValue = newValue;
                that.setAttribute('data-value', newValue);
                that._refreshSelectedIndex();
            }
            else {
                const value = newValue.join(',');
                const oldValue = that.value;
                that.$.input.dataValue = value;
                that._setSelection();

                that.propertyChangedHandler('value', oldValue, that.value);
            }
        }
        else if (propertyName === 'value') {
            if (that.displayMember !== that.valueMember && typeof newValue !== 'string') {
                that.set('value', newValue[that.displayMember]);
                that.$.input.dataValue = newValue[that.valueMember];
                that.setAttribute('data-value', that.$.input.dataValue);
            }
            else {
                that.$.input.value = that.$.input.dataValue = newValue;
                that.setAttribute('data-value', newValue);
            }

            that._refreshSelectedIndex();
        }
        else if (propertyName === 'readonly') {
            that._setAriaRelations();
            that._setInputPurpose();
        }
        else if (propertyName === 'theme' || propertyName === 'rightToLeft') {
            that.$.scrollView[propertyName] = newValue;
        }
        else if (propertyName === 'inputPurpose') {
            that._setInputPurpose();
        }
        else if (propertyName === 'inverted') {
            newValue ? that.$.scrollView.setAttribute('inverted', '') : that.$.scrollView.removeAttribute('inverted');
        }
        else if (propertyName === 'disabled' || propertyName === 'unfocusable') {
            that._setFocusable();
        }
    }

    _updateTargetValue() {
        const that = this;
        if (that.dataset['target']) {
            const target = document.querySelector(that.dataset['target']);

            if (target) {
                const property = that.dataset['property'];
                const value = that.value;

                if (property && target[property] !== undefined) {
                    target[property] = value;
                }
            }
        }
    }

    /**
     * Called when the element is detached from the DOM.
     */
    detached() {
        const that = this;

        super.detached();
        that.close();
        that.removeAttribute('focus');
    }

    _getDataSourceFromOption() {
        const that = this;
        if (that.innerHTML.indexOf('<option') >= 0) {
            that.$.optionsContainer.innerHTML = that.innerHTML.trim();
            const options = that.querySelectorAll('option');
            let dataSource = [];
            for (let i = 0; i < options.length; i++) {
                const option = options[i];
                const selected = option.hasAttribute('selected') || (option.hasAttribute('is-selected') && option.getAttribute('is-selected') !== 'false');
                let label = ('' + option.textContent).trim();
                const menuLabel = ('' + option.innerHTML).trim();
                const value = option.hasAttribute('value') ? option.getAttribute('value') : label;
                const disabled = option.hasAttribute('disabled');
                const labelAttribute = option.hasAttribute('label') ? option.getAttribute('label') : null;

                if (labelAttribute) {
                    label = labelAttribute;
                }

                dataSource.push({
                    label: label,
                    menuLabel: menuLabel,
                    value: value,
                    disabled: disabled,
                    selected: selected,
                    className: option.className,
                    dataset: option.dataset
                });

                option.classList.add('smart-hidden');
            }
            that.dataSource = dataSource;
        }
    }
    render() {
        const that = this,
            dropDownClassList = that.dropDownClassList;


        that._getDataSourceFromOption();

        if (that.selectedValues && that.selectedValues.length) {
            const value = that.selectedValues.join(',');
            that.value = value;
        }

        if (that._createElement) {
            that._createElement();
        }

        const scrollView = that.$.scrollView;

        for (let i = 0, max = dropDownClassList.length; i < max; i += 1) {
            scrollView.classList.add(dropDownClassList[i]);
        }

        if (that.opened) {
            that.opened = false;
            that.open();
        }

        const menu = that.$.menu;

        if (menu) {
            menu.setAttribute('role', 'presentation');
            scrollView.appendChild(menu);
        }

        if (that.dropDownHeight !== 'auto') {
            scrollView.style.setProperty('--smart-input-drop-down-menu-height', that.dropDownHeight + 'px');
        }
        else {
            scrollView.style.setProperty('--smart-input-drop-down-menu-height', 'auto');
        }

        scrollView.classList.add('smart-input-drop-down-menu');

        for (let i = 0; i < that.classList.length; i++) {
            const className = that.classList[i];

            if (className === 'smart-element' || className === 'smart-input') {
                continue;
            }

            scrollView.classList.add(className + '-drop-down-menu');
        }

        scrollView.id = that.id + '_' + that.tagName.toLowerCase() + '_menu_' + Math.floor((1 + Math.random()) * 0x10000).toString(16).substring(1);
        scrollView.ownerElement = that;

        let items = that.dataSource;

        if (typeof items === 'string') {
            items = that.$.deserialize(items, 'array');
        }

        if (items && items.length && Array.isArray(items)) {
            let selectedIndex = -1;

            items.forEach((item, index) => {
                if (item.selected) {
                    selectedIndex = index;
                }
            });

            if (selectedIndex >= 0) {
                that.selectedIndex = selectedIndex;

                const item = items[selectedIndex];

                if (typeof item === 'string') {
                    that.selectedValue = items[selectedIndex];
                }
                else {
                    that.selectedValue = items[selectedIndex].value;
                }

            }
        }

        that._setAriaRelations();
        that._setInputPurpose();
        that._setSelection();
        that.isInitialized = true;

        super.render();
    }

    _setSelection(noEvent) {
        const that = this;
        let items = that.dataSource;

        if (typeof items === 'string') {
            items = that.$.deserialize(items, 'array');
        }

        if (that.selectedIndex >= 0 && items && items.length > 0) {
            let item = items[that.selectedIndex];

            if (typeof item === 'number' || typeof item === 'boolean') {
                item = '' + item;
            }

            if (typeof item === 'string') {
                let oldValue = that.$.input.dataValue;
                let oldLabel = that.value;

                that.set('value', item);
                that.$.input.dataValue = item;
                that.$.input.setAttribute('data-value', item);
                that.selectedValues = [that.$.input.dataValue];

                if (that.isInitialized && noEvent !== false) {
                    that.$.fireEvent('change', { value: that.value, label: that.label, oldValue: oldValue, oldLabel: oldLabel });
                    that._updateTargetValue();
                }
            }
            else {
                let oldValue = that.$.input.dataValue;
                let oldLabel = that.value;

                if (item) {
                    if (item.value !== undefined) {
                        that.set('value', item.value !== undefined ? item.value + '' : '');
                        that.$.input.dataValue = item.value;
                        that.$.input.setAttribute('data-value', item.value);
                        that.selectedValues = [that.$.input.dataValue];
                    }
                    else if (item.label) {
                        that.selectedValues = [item.value];
                    }

                    if (item.label) {
                        that.set('value', item.label);
                    }
                }

                if (that.isInitialized && noEvent !== false) {
                    that.$.fireEvent('change', { value: that.$.input.dataValue, label: item.value, oldValue: oldValue, oldLabel: oldLabel });
                    that._updateTargetValue();
                }
            }
        }
    }

    _createElement() {
        const that = this;
        const menu = document.createElement('ul');
        const scrollView = document.createElement('smart-scroll-viewer');

        scrollView.rightToLeft = that.rightToLeft;
        scrollView.theme = that.theme;
        scrollView.horizontalScrollBarVisibility = 'hidden';

        that.inverted ?
            scrollView.setAttribute('inverted', '') :
            scrollView.removeAttribute('inverted');

        that.$.scrollView = scrollView;
        that.$.menu = menu;

        that._setFocusable();

        const input = that.$.input;
        //that.$.inputContainer.tabIndex = 0;
        that.$.menu.onclick = function (event) {
            const li = event.target.closest('li');

            event.stopPropagation()
            event.preventDefault();

            that._performSelect();

            if (li) {
                that.$.fireEvent('itemClick', { index: [...li.parentNode.children].indexOf(li), item: li, label: li.getAttribute('aria-label'), value: li.dataValue });
            }

            input.focus();
        }
    }

    /**
     * Sets WAI-ARIA relations.
     */
    _setAriaRelations() {
        const that = this,
            label = that.getAttribute('aria-label');

        if (that.readonly) {
            that.setAttribute('role', 'button');

            if (!label && that.placeholder) {
                that.setAttribute('aria-label', that.placeholder);
            }

            that.removeAttribute('aria-readonly');
            that.$.input.setAttribute('aria-hidden', true);
            that.$.input.removeAttribute('aria-activedescendant');
            that.$.input.removeAttribute('aria-controls');
            that.$.dropDownButton.setAttribute('aria-hidden', true);
        }
        else {
            that.setAttribute('role', 'combobox');

            if (label && label === that.placeholder) {
                that.removeAttribute('aria-label');
            }

            that.$.input.setAttribute('role', 'searchbox');
            that.$.input.removeAttribute('aria-hidden', true);
            that.$.dropDownButton.removeAttribute('aria-hidden');
        }

        that.setAttribute('aria-expanded', that.opened);
        that.setAttribute('aria-haspopup', 'listbox');
        that.setAttribute('aria-owns', that.$.scrollView.id);
        that.$.scrollView.setAttribute('role', 'listbox');
    }

    /**
     * Sets WAI-ARIA property aria-activedescendant.
     */
    _setActiveDescendant(item) {
        const that = this;

        if (!that.readonly) {
            if (item) {
                that.setAttribute('aria-activedescendant', item.id);
            }
            else {
                that.removeAttribute('aria-activedescendant');
            }
        }
    }

    /**
     * Sets the HTML autocomplete attribute based on the property "inputPurpose".
     */
    _setInputPurpose() {
        const that = this;

        if (that.readonly) {
            that.$.input.removeAttribute('autocomplete');
        }
        else {
            that.$.input.setAttribute('autocomplete', that.inputPurpose);
        }
    }

    /**
     * Sets whether the element can be focused.
     */
    _setFocusable() {
        const that = this;

        if (that.disabled || that.unfocusable) {
            that.$.input.setAttribute('tabindex', -1);
        }
        else {
            that.$.input.removeAttribute('tabindex');
        }
    }
});
